from tethysts import Tethys
import xarray as xr
import yaml
import pandas as pd
import os
import shapely
import geopandas as gpd
from shapely.geometry import shape
from shapely.strtree import STRtree
import pickle
from shapely import wkb, wkt
import orjson
import tethys_utils as tu
import numpy as np

pd.options.display.max_columns = 10

##############################################
### Parameters

base_dir = os.path.split(os.path.split(os.path.realpath(os.path.dirname(__file__)))[0])[0]

with open(os.path.join(base_dir, 'parameters.yml')) as param:
    param = yaml.safe_load(param)

remotes_list = param['remotes']

######################################
### Testing

remote = {'bucket': 'fire-emergency-nz', 'connection_config': 'https://b2.tethys-ts.xyz', 'version': 2}
remote = {'bucket': 'es-hilltop', 'connection_config': 'https://b2.tethys-ts.xyz'}
remote = {'bucket': 'niwa-cliflo', 'connection_config': 'https://b2.tethys-ts.xyz'}
remote = {'bucket': 'ecan-env-monitoring', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 2}
remote = {'bucket': 'point-forecasts', 'connection_config': 'https://b2.tethys-ts.xyz', 'version': 3}
remote = {'bucket': 'met-solutions', 'connection_config': 'https://b2.tethys-ts.xyz', 'version': 3}
# remote = {'bucket': 'nasa-data', 'connection_config': 'https://b2.tethys-ts.xyz'}
remote = {'bucket': 'tethysts', 'connection_config': 'https://b2.tethys-ts.xyz', 'version': 3}
remote = {'bucket': 'nz-open-modelling-consortium', 'connection_config': 'https://b2.nzrivers.xyz', 'version': 3}
remote = {'bucket': 'typhon', 'public_url': 'https://b2.tethys-ts.xyz', 'version': 4}

remote = remotes_list[-1]
# remote['connection_config'] = 'https://b2.tethys-ts.xyz'
#
# dataset_id = '4edc94c19bf074027bc7c099'
# dataset_id = 'dddb02cd5cb7ae191311ab19'
# station_id = 'fedeb59e6c7f47597a7d47c7'
# station_id = 'fe9a63fae6f7fe58474bb3c0'
# station_id2 = 'fe9a63fae6f7fe58474bb3c0'
# station_id = '6b75a7fb1453ef94148bda19'
# station_ids = [station_id, '5d06c5a8065a26b51c19b241']
dataset_id = 'dddb02cd5cb7ae191311ab19'
station_ids = ['fedeb59e6c7f47597a7d47c7', 'fe9a63fae6f7fe58474bb3c0']

dataset_id='0b2bd62cc42f3096136f11e9'
station_id='e0c38cc6fd3eb51fb553d45a'

dataset_id = 'de3bff8e3c3a2ad9200d8684'
station_id = 'fedeb59e6c7f47597a7d47c7'

# dataset_id='320d6836250169a5f7b78163'
# station_id='7df0d7fe8c6fcd06c50d73a6'
station_id = 'de7cfd749e24a6b78c2281fb'
station_ids = [station_id, 'de7cfd749e24a6b78c2281fb']
dataset_id = '4690821a39b3fb65c197a540'
station_id = '78b7a3d3f61cf33e681490c1'
station_ids = [station_id, '9f3ee54af239b845241f8f13']
dataset_id = '38138ea1c3350031d1b217f6'
station_id = 'b318207aa246e7bbbd74cb19'

dataset_id = 'f56892eb59d12cfbc02acceb'
station_id = 'f4da7cc7f4c947b2e6b30344'

dataset_id = 'fb77f37b16edae3534e73ddd'
station_id = 'fd60b5bafe19d6b243dda43d'

dataset_id = 'e37f1451fcf8f9e64b66be8d'
7300
dataset_id = '9c7e107f99180e45eafdf5af'
station_id = 'c6df8b47b2efce3daedef48e'
station_id = 'a3f2d35fd6df8247cea32d03'

dataset_id = '9bf36a9e6b6a2a111bf6634b'
station_id = '4db28a9db0cb036507490887'

dataset_id = '9845cd0049891916f2a59c80'
station_id = '02d4943e784fcb6acd819b72'

dataset_id = '2c004d8366bcc22927d68994'
station_id = 'fe67563f37772e63c74487be'
station_ids = [station_id, 'fe35e6509703baebf294c59e']

dataset_id = '2c004d8366bcc22927d68994'
station_id = '673d6d9fca3ccf38fa009ad1'

dataset_id = 'f27574a7b38eab5b0bc2b3d7' # envlib
station_id = 'fedeb59e6c7f47597tasman-enva7d47c7'

dataset_id = 'de3bff8e3c3a2ad9200d8684'
station_id = 'fedeb59e6c7f47597a7d47c7'

dataset_id = '99d5109c61642e55e8a57657'
station_id = 'de3f6e8951378d6c16186b8f'

dataset_id = 'f27574a7b38eab5b0bc2b3d7'
station_id = '9c90243e84b8c5b17f0726c4'

dataset_id = '0a1583c61202c6791ae39e63'
station_id = '5d06c5a8065a26b51c19b241'

dataset_id = '0a1583c61202c6791ae39e63'

dataset_id = '692e8696cc9e11b4a1ef943d'

dataset_id = '799329b9023c9d9980abd1f6'
station_id = '937f2bb85e82347746fe1be9'

dataset_id = '4f5945540c2391967b550cc6'
station_id = '4db28a9db0cb036507490887'

dataset_id = '52e2196ce75eba1b79e61680'
station_id = '7ddd130ed4b89381d879e0d5'
#
#
self = Tethys([remote])
self = Tethys(remotes_list)
#
stn_list1 = self.get_stations(dataset_id)
stn_list1 = self.get_stations(dataset_id, results_object_keys=True)
# run_dates = self.get_run_dates(dataset_id, station_id)
data1 = self.get_results(dataset_id, station_id, output='Dataset')
data1 = self.get_results(dataset_id, station_id, output='Dataset', cache='memory')
# data1 = self.get_results(dataset_id, station_id, modified_date=True, quality_code=True, remove_height=True, output='DataArray')
# data1 = self.get_results(dataset_id, station_id, modified_date=True, quality_code=True, remove_height=True, output='DataArray')
# data1 = self.get_results(dataset_id, station_id, modified_date=True, quality_code=True, output='Dict')
# data1 = self.get_results(dataset_id, station_id, output='Dict')
data1 = self.get_results(dataset_id, station_id, run_date='2021-08-29T00:00:00', output='Dataset')
data1 = self.get_results(dataset_id, station_id, run_date='2021-03-12T18:00:16', output='Dataset')

data1 = self.get_results(dataset_id, station_id, squeeze_dims=True, output='DataArray', cache='memory')
data2 = self.get_results(dataset_id, station_id2, output='DataArray')

run_dates1 = self.get_run_dates(dataset_id, station_id)

station_ids = [s['station_id'] for s in stn_list1[111:160]]
data2 = self.get_bulk_results(dataset_id, station_ids, output='Dataset', threads=10)

station_ids = [s['station_id'] for s in stn_list1[100:130]]
data3 = self.get_bulk_results(dataset_id, station_ids, output='Dataset', threads=10)

# dataset_id = 'f4cfb5a362707785dd39ff85'
# station_id = 'ff4213c61878e098e07df513'

query_geometry = {'type': 'Point', 'coordinates': [171.043868, -43.106372]}
geom_query = shape(query_geometry).buffer(0.1)

stn_list2 = self.get_stations(dataset_id, query_geometry)
stn_list2 = self.get_stations(dataset_id, lat=-43.1, lon=171.1)
data2 = self.get_results(dataset_id, geometry=query_geometry, output='Dataset')

stn = [s for s in stn_list1 if s['station_id'] == station_id]

stn = [s for s in stn_list1 if 'Waiau River' in s['ref']]


gwl_ds1 = [d for d in self.datasets if d['parameter'] == 'groundwater_depth']
era5_ds1 = [d for d in self.datasets if d['owner'] == 'ECMWF']
nz_ds1 = [d for d in self.datasets if d['owner'] == 'NZ Open Modelling Consortium']
hh_ds1 = [d for d in self.datasets if d['owner'] == 'Headwaters Hydrology'][0]

stns = self.get_stations(hh_ds1['dataset_id'])

stn_id = 'fd60b5bafe19d6b243dda43d'

data1 = self.get_results(hh_ds1['dataset_id'], stn_id, output='Dataset')


type1 = stns[0]['geometry']['type']

if type1 == 'Polygon':
    geo = [s['geometry']['coordinates'][0][0][-1] for s in stns]
    geo1 = np.array(geo).round(5)
    geo1.sort()
    np.diff(geo1)


# geo1 = []
# for s in stn_list1:
#     s1 = {'station_id': s['station_id'], 'lon': s['geometry']['coordinates'][0], 'lat': s['geometry']['coordinates'][1]}
#     geo1.append(s1)

geo1 = []
for s in stn_list1:
    s1 = {'station_id': s['station_id'], 'geo': shape(s['geometry']).simplify(0.1)}
    geo1.append(s1)

sdf = pd.DataFrame(geo1)

sdf2 = gpd.GeoDataFrame(sdf[['station_id']], geometry=sdf['geo'])
sdf2['x'] = sdf2.geometry.x.round(1)
sdf2['y'] = sdf2.geometry.y.round(1)

#####################################################
## dev testing

geometry: dict = None
lat: float = None
lon: float = None
from_date = None
to_date = None
from_mod_date = None
to_mod_date = None
version_date = None
heights = None
bands: int = None
squeeze_dims: bool = False
threads: int = 30
output_path = None
compression='zstd'

remote = {'bucket': 'nz-open-modelling-consortium', 'public_url': 'https://b2.nzrivers.xyz/file/', 'version': 4}
remote = {'bucket': 'fire-emergency-nz', 'public_url': 'https://b2.tethys-ts.xyz/file/', 'version': 4}
remote = {'bucket': 'nasa-data', 'public_url': 'https://b2.tethys-ts.xyz/file/', 'version': 4}
remote = {'bucket': 'ecan-env-monitoring', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4}
remote = {'bucket': 'gwrc-env', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4}

remote = {'bucket': 'typhon', 'public_url': 'https://b2.tethys-ts.xyz/file/', 'version': 4}
remote = {'bucket': 'point-forecasts', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4}
remote = {'bucket': 'es-hilltop', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4}
remote = {'bucket': 'nz-water-use', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4}
remote = {'bucket': 'niwa-sos', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4}
remote = {'bucket': 'orc-env', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4}
remote = {'bucket': 'met-service', 'public_url': 'https://b2.nzrivers.xyz/file/', 'version': 4}
remote = {'bucket': 'met-solutions', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4}
remote = {'bucket': 'tasman-env', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4}
remote = {'bucket': 'noaa-nwm', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4}
remote = {'bucket': 'jaxa-data', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4}
remote = {'bucket': 'linz-data', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4}
remote = {'bucket': 'mdc-env', 'public_url': 'https://b2.tethys-ts.xyz/file', 'version': 4}

cache = '/media/nvme1/cache/tethys'
# cache = '/home/mike/cache/tethys'

dataset_id = '7751c5f1bf47867fb109d7eb'
dataset_id = '0b2bd62cc42f3096136f11e9'
dataset_id = 'fb60aaa921d35a33727b53fe'
dataset_id = 'f16774ea29f024a306c7fc7a'
dataset_id = '9568f663d566aabb62a8e98e'



dataset_id = '4ae05d099af292fec48792ec'
station_ids = '462fe69802f7df152eef7651'

dataset_id = '7751c5f1bf47867fb109d7eb'

dataset_id = 'b5d84aa773de2a747079c127'
station_ids = 'f9c61373e7ca386c1fab06db'
station_ids = ['f9c61373e7ca386c1fab06db', '96e9ff9437fc738b24d10b42']

dataset_id = 'dddb02cd5cb7ae191311ab19'
station_ids = '71369f685f7a5841a060a171'

dataset_id = '0b2bd62cc42f3096136f11e9'
station_ids = 'c8db6013a9eb76705b5c80f2'
ref = 'ashley'
station_ids = '7eb2694917a2ece89e86e9b8'

dataset_id = 'b3d852cd72ac043c701493c4'

station_ids = 'e8ef98448a4cd1df6d6c6af3'

dataset_id = '9568f663d566aabb62a8e98e'

dataset_id = 'f56892eb59d12cfbc02acceb'
version_date='2022-04-29T12:05:04'

dataset_id = 'c3d8d28896ec5f44d8e46f43'

dataset_id = '752ee66d969cc09a16efebc7'
dataset_id = 'a746aca4bd66981f148e680f'

dataset_id = '10c4bea7d538387033956cf8'

dataset_id = 'd3e5979d76fcebc9f6aec383'
version_date = '2022-09-04T19:00:00'

dataset_id = '469b6a9ef620bce70fab5760'
dataset_id = '980177538186a9bb9c9a0672'
dataset_id = '54374801c0311a98a0f8e5ef'

dataset_id = 'ef79659d4175c6d70d748b5e'
station_ids = 'b8db7f021cf6e422d6d85881'

self = Tethys([remote], cache=cache)
self = Tethys([remote])
self = Tethys()

rv1 = self.get_versions(dataset_id)
stns1 = self.get_stations(dataset_id)
stns1 = self.get_stations(dataset_id, version_date=version_date)

station_ids = [s['station_id'] for s in stns1[:2]]
station_ids = [s['station_id'] for s in stns1 if ref in s['ref']]

results1 = self.get_results(dataset_id, station_ids, heights=None)
results1 = self.get_results(dataset_id, station_ids, heights=None, output_path='/media/nvme1/cache/tethys/test.h5', compression='zstd')

results1 = self.get_results(dataset_id, station_ids, heights=[10, 20, 30, 80], version_date=version_date)

results1 = self.get_results(dataset_id, station_ids, heights=[10])
results1 = self.get_results(dataset_id, station_ids, heights=[10], output_path='/media/nvme1/cache/tethys/test.h5', compression='zstd')

results1 = self.get_results(dataset_id, station_ids, heights=[10], from_date='2020-04-01')

results1 = self.get_results(dataset_id, station_ids, heights=None, from_date='2015-04-01')

results1 = self.get_results(dataset_id, station_ids, heights=None, from_mod_date='2022-10-01')
results1 = self.get_results(dataset_id, station_ids, heights=None, to_mod_date='2022-10-01')


for d in self.datasets:
    rv1 = self.get_versions(d['dataset_id'])
    print(rv1)


results1 = self.get_results(dataset_id, 'fac7e3f6ee48113ccb30e446', heights=None)

results2 = self.get_results(dataset_id, station_ids, heights=[10], version_date='2022-03-01')


station_ids = [s['station_id'] for s in stns1[:3]]

for s in stns1:
    print(s['station_id'])
    out1 = self.get_results(dataset_id, s['station_id'])



### Tests
remote1 = {'bucket': 'fire-emergency-nz', 'public_url': 'https://b2.tethys-ts.xyz/file/', 'version': 4}
remote2 = {'bucket': 'fire-emergency-nz', 'public_url': 'https://b2.tethys-ts.xyz/file/', 'version': 2}

ds_ids = ['c5f55d97e71e7cd73295ad7f', 'dddb02cd5cb7ae191311ab19', '0b2bd62cc42f3096136f11e9']

t1 = Tethys([remote1])
t2 = Tethys([remote2])

for ds_id in ds_ids:
    print('-- ds_id: ' + ds_id)
    stns = t1.get_stations(ds_id)

    stns0 = t2.get_stations(ds_id)

    for stn in stns:
        stn_id = stn['station_id']
        print('-- stn_id: ' + stn_id)
        base_ref = stn['ref']

        if base_ref[-3] == '_':
            old_ref = base_ref[:-3]
        else:
            old_ref = base_ref

        stn0 = [s for s in stns0 if old_ref in s['ref']]

        if stn0:
            stn0 = stn0[0]
            old_stn_id = stn0['station_id']

            r1 = t1.get_results(ds_id, stn_id)
            r1_prod = int(np.prod(list(dict(r1.dims).values())))

            r2 = t2.get_results(ds_id, old_stn_id)
            r2_prod = int(np.prod(list(dict(r2.dims).values())))

            if r1_prod < r2_prod:
                print('{ds_id}: old stn {old_stn_id}: {old_prod}, new stn {new_stn_id}: {new_prod}'.format(ds_id=ds_id, old_stn_id=old_stn_id, new_stn_id=stn_id, old_prod=r2_prod, new_prod=r1_prod))



stn_ids = {s['station_id']: pd.Timestamp(s['time_range']['to_date']) - pd.Timestamp(s['time_range']['from_date']) for s in stns1}




test_stn_dict = {'c5f55d97e71e7cd73295ad7f': [['91d7d738684b3b6d8e11acd9', '276a3e6428a700229b437626'], ['17c7c90057683b807ad77b10', '751946ea52d04e67639fea1c']],
                 'dddb02cd5cb7ae191311ab19': [['91d7d738684b3b6d8e11acd9', '276a3e6428a700229b437626'], ['17c7c90057683b807ad77b10', '751946ea52d04e67639fea1c']],
                 '0b2bd62cc42f3096136f11e9': [['91d7d738684b3b6d8e11acd9', '276a3e6428a700229b437626'], ['71cd89d47beb79712903eb10', '71369f685f7a5841a060a171'], ['17c7c90057683b807ad77b10', '751946ea52d04e67639fea1c']]}


ds_id = 'c5f55d97e71e7cd73295ad7f'


stn = [s for s in stns if s['station_id'] == '751946ea52d04e67639fea1c'][0]
stn0 = [s for s in stns0 if s['station_id'] == '17c7c90057683b807ad77b10'][0]

[s for s in stns1 if 'Waimak' in s['name']]

[s for s in stns0 if 'poroporo' in s['ref']]


ds1 = [d for d in self.datasets if (d['owner'] == 'NZ Open Modelling Consortium') and (d['product_code'] == 'UC WRF NZ South Island Marlborough Nelson 1km')]

ds1 = [d for d in self.datasets if (d['feature'] == 'atmosphere')]

ds1 = [d for d in self.datasets if (d['owner'] == 'MetService')]

ds1 = [d for d in self.datasets if (d['feature'] == 'waterway')]



import os
from tethysts import Tethys
from pprint import pprint
import numpy as np
from shapely.geometry import box, mapping

###########################################
### Parameters

dataset_id = '469b6a9ef620bce70fab5760'
cache = '/home/mike/cache/tethys'
output_path = '/home/mike/cache/temp'

heights = [10, 20] # Select the heights you want to extract
block_length = 2   # Set this to at least 2. You can set it to 3 if you've got enough RAM


##########################################
### Get the data

# Cache the results chunks so that it's easy to query the results later
ts = Tethys(cache=cache)

ds1 = [d for d in self.datasets if (d['dataset_id'] == dataset_id)][0]

pprint(ds1) # Take a look at the dataset metadata

# Get the total extent of the dataset and break it up by the block_length
extent = np.array(ds1['extent']['coordinates'][0])
extent_transposed = extent.T

min_lon = extent_transposed[0].min()
max_lon = extent_transposed[0].max()
min_lat = extent_transposed[1].min()
max_lat = extent_transposed[1].max()

x_range = np.arange(min_lon, max_lon + block_length, block_length)
y_range = np.arange(min_lat, max_lat + block_length, block_length)

# Iterate through all combos of lons and lats
for iy, y in enumerate(y_range[1:]):
    min_y = y_range[iy]
    for ix, x in enumerate(x_range[1:]):
        min_x = x_range[ix]

        new_extent = box(min_x, min_y, x, y)
        geo = mapping(new_extent)

        stns = ts.get_stations(dataset_id, geo)
        stn_ids = [s['station_id'] for s in stns]

        # If stations are found, then get the associated results
        if stn_ids:
            results = ts.get_results(dataset_id, stn_ids, heights=heights)

            output_file = '{ds}_{min_x}_{min_y}_{max_x}_{max_y}.nc'.format(ds=dataset_id, min_x=min_x, min_y=min_y, max_x=x, max_y=y)

            # Save the file to netcdf (xarray uses an excessive amount of memory when saving to netcdf, so be careful...)
            results.to_netcdf(os.path.join(output_path, output_file))
            results.close()
            del results





























############################3
### checks

base_path = '/media/nvme1/cache/tethys/preprocessed_data'


for i in range(1000000):
    print(i)
    check_compressor(obj)

for i in range(1000000):
    print(i)
    check_writing(obj, base_path)


with concurrent.futures.ProcessPoolExecutor(max_workers=3, mp_context=mp.get_context("spawn")) as executor:
    futures = []
    for i in range(1000000):
        f = executor.submit(tu.misc.check_compressor, obj)
        futures.append(f)
    runs = concurrent.futures.wait(futures)

chunks_list = [r.result() for r in runs[0]]


with concurrent.futures.ProcessPoolExecutor(max_workers=3, mp_context=mp.get_context("spawn")) as executor:
    futures = []
    for i in range(1000000):
        f = executor.submit(tu.misc.check_writing, obj, base_path)
        futures.append(f)
    runs = concurrent.futures.wait(futures)

chunks_list = [r.result() for r in runs[0]]




remotes = [{'bucket': 'ecan-env-monitoring', 'public_url': 'https://b2.tethys-ts.xyz/file/', 'version': 4}]
dataset_id = 'b5d84aa773de2a747079c127'
station_id = 'f9c61373e7ca386c1fab06db'
geometry = {'type': 'Point', 'coordinates': [172.0, -42.8]}
lon = 172.0
lat = -42.8
distance = 0.2


ts = Tethys()
datasets = ts.datasets
my_dataset = [d for d in datasets if (d['parameter'] == 'precipitation') and
                                     (d['product_code'] == 'raw_data') and
                                     (d['frequency_interval'] == '24H') and
                                     (d['owner'] == 'Environment Canterbury')][0]


stations = ts.get_stations(dataset_id)
my_station = [s for s in stations if (s['name'] == "Waimakariri at Arthur's Pass")][0]
my_station = ts.get_stations(dataset_id, geometry=geometry)
my_stations = ts.get_stations(dataset_id, lat=lat, lon=lon, distance=distance)
results = ts.get_results(dataset_id, station_id, output='xarray')


















































"""


"""
import io
import os
import numpy as np
# import requests
import xarray as xr
import pandas as pd
import orjson
from datetime import datetime
import zstandard as zstd
import pickle
import copy
# import boto3
import botocore
from time import sleep
from shapely.geometry import shape, Polygon, Point
from shapely.strtree import STRtree
from typing import Optional, List, Any, Union
from scipy import spatial
# import tethys_data_models as tdm
import pathlib
# from functools import partial
from pydantic import HttpUrl
# import shutil
# import gzip
from hdf5tools import H5
import s3tethys
# import smart_open
# import psutil
from pprint import pprint

pd.options.display.max_columns = 10

##############################################
### Reference objects

b2_public_key_pattern = '{base_url}/{bucket}/{obj_key}'
contabo_public_key_pattern = '{base_url}:{bucket}/{obj_key}'
public_remote_key = 'https://b2.tethys-ts.xyz/file/tethysts/tethys/public_remotes_v4.json.zst'

local_results_name = '{ds_id}/{version_date}/{stn_id}/{chunk_id}.{chunk_hash}.results.h5'

s3_url_base = 's3://{bucket}/{key}'

##############################################
### Helper functions


def update_nested(in_dict, ds_id, version_date, value):
    """

    """
    if ds_id in in_dict:
        in_dict[ds_id][version_date] = value
    else:
        in_dict.update({ds_id: {version_date: value}})


def make_run_date_key(run_date=None):
    """

    """
    if run_date is None:
        run_date = pd.Timestamp.today(tz='utc')
        run_date_key = run_date.strftime('%Y%m%dT%H%M%SZ')
    elif isinstance(run_date, pd.Timestamp):
        run_date_key = run_date.strftime('%Y%m%dT%H%M%SZ')
    elif isinstance(run_date, str):
        run_date1 = pd.Timestamp(run_date).tz_localize(None)
        run_date_key = run_date1.strftime('%Y%m%dT%H%M%SZ')
    else:
        raise TypeError('run_date must be None, Timestamp, or a string representation of a timestamp')

    return run_date_key


def create_public_s3_url(base_url, bucket, obj_key):
    """
    This should be updated as more S3 providers are added!
    """
    if 'contabo' in base_url:
        key = contabo_public_key_pattern.format(base_url=base_url.rstrip('/'), bucket=bucket, obj_key=obj_key)
    else:
        key = b2_public_key_pattern.format(base_url=base_url.rstrip('/'), bucket=bucket, obj_key=obj_key)

    return key


# class ResponseStream(object):
#     """
#     In many applications, you'd like to access a requests response as a file-like object, simply having .read(), .seek(), and .tell() as normal. Especially when you only want to partially download a file, it'd be extra convenient if you could use a normal file interface for it, loading as needed.

# This is a wrapper class for doing that. Only bytes you request will be loaded - see the example in the gist itself.

# https://gist.github.com/obskyr/b9d4b4223e7eaf4eedcd9defabb34f13
#     """
#     def __init__(self, request_iterator):
#         self._bytes = BytesIO()
#         self._iterator = request_iterator

#     def _load_all(self):
#         self._bytes.seek(0, SEEK_END)
#         for chunk in self._iterator:
#             self._bytes.write(chunk)

#     def _load_until(self, goal_position):
#         current_position = self._bytes.seek(0, SEEK_END)
#         while current_position < goal_position:
#             try:
#                 current_position += self._bytes.write(next(self._iterator))
#             except StopIteration:
#                 break

#     def tell(self):
#         return self._bytes.tell()

#     def read(self, size=None):
#         left_off_at = self._bytes.tell()
#         if size is None:
#             self._load_all()
#         else:
#             goal_position = left_off_at + size
#             self._load_until(goal_position)

#         self._bytes.seek(left_off_at)
#         return self._bytes.read(size)

#     def seek(self, position, whence=SEEK_SET):
#         if whence == SEEK_END:
#             self._load_all()
#         else:
#             self._bytes.seek(position, whence)


def cartesian_product(*arrays):
    la = len(arrays)
    dtype = np.result_type(*arrays)
    arr = np.empty([len(a) for a in arrays] + [la], dtype=dtype)
    for i, a in enumerate(np.ix_(*arrays)):
        arr[...,i] = a
    return arr.reshape(-1, la)


def get_nearest_station(stns, geom_query):
    """

    """
    if isinstance(geom_query, dict):
        geom_query = shape(geom_query)

    geom1 = [shape(s['geometry']) for i, s in stns.items()]
    strtree = STRtree(geom1)
    res = strtree.nearest(geom_query)
    res_id = res.wkb_hex

    stn_id_dict = {shape(s['geometry']).wkb_hex: i for i, s in stns.items()}

    stn_id = stn_id_dict[res_id]

    return stn_id


def get_intersected_stations(stns, geom_query):
    """

    """
    if isinstance(geom_query, dict):
        geom_query = shape(geom_query)

    geom1 = [shape(s['geometry']) for i, s in stns.items()]
    strtree = STRtree(geom1)
    res = strtree.query(geom_query)
    res_ids = [r.wkb_hex for r in res]

    stn_id_dict = {shape(s['geometry']).wkb_hex: i for i, s in stns.items()}

    stn_ids = [stn_id_dict[r] for r in res_ids]

    return stn_ids


def spatial_query(stns: dict,
                  query_geometry: Optional[dict] = None,
                  lat: Optional[float] = None,
                  lon: Optional[float] = None,
                  distance: Optional[float] = None):
    """

    """
    if isinstance(lat, float) and isinstance(lon, float):
        geom_query = Point(lon, lat)
        if isinstance(distance, (int, float)):
            geom_query = geom_query.buffer(distance)
            stn_ids = get_intersected_stations(stns, geom_query)
        else:
            stn_ids = [get_nearest_station(stns, geom_query)]
    elif isinstance(query_geometry, dict):
        geom_query = shape(query_geometry)
        if isinstance(geom_query, Point):
            stn_ids = [get_nearest_station(stns, geom_query)]
        elif isinstance(geom_query, Polygon):
            stn_ids = get_intersected_stations(stns, geom_query)
        else:
            raise ValueError('query_geometry must be a Point or Polygon dict.')
    else:
        stn_ids = None

    return stn_ids


def get_nearest_from_extent(data,
                            query_geometry: Optional[dict] = None,
                            lat: Optional[float] = None,
                            lon: Optional[float] = None):
    """

    """
    ## Prep the query point
    if isinstance(lat, float) and isinstance(lon, float):
        geom_query = Point(lon, lat)
    elif isinstance(query_geometry, dict):
        geom_query = shape(query_geometry)
        if not isinstance(geom_query, Point):
            raise ValueError('query_geometry must be a Point.')
    else:
        raise ValueError('query_geometry or lat/lon must be passed as a Point.')

    ## Prep the input data
    if 'geometry' in data:
        raise NotImplementedError('Need to implement geometry blocks nearest query.')
    else:
        lats = data['lat'].values
        lons = data['lon'].values
        xy = cartesian_product(lons, lats)
        kdtree = spatial.cKDTree(xy)
        dist, index = kdtree.query(geom_query.coords[0])
        lon_n, lat_n = xy[index]

    data1 = data.sel(lon=[lon_n], lat=[lat_n])

    return data1


def read_pkl_zstd(obj, unpickle=False):
    """
    Deserializer from a pickled object compressed with zstandard.

    Parameters
    ----------
    obj : bytes or str
        Either a bytes object that has been pickled and compressed or a str path to the file object.
    unpickle : bool
        Should the bytes object be unpickled or left as bytes?

    Returns
    -------
    Python object
    """
    if isinstance(obj, str):
        with open(obj, 'rb') as p:
            dctx = zstd.ZstdDecompressor()
            with dctx.stream_reader(p) as reader:
                obj1 = reader.read()

    elif isinstance(obj, bytes):
        dctx = zstd.ZstdDecompressor()
        obj1 = dctx.decompress(obj)
    else:
        raise TypeError('obj must either be a str path or a bytes object')

    if unpickle:
        obj1 = pickle.loads(obj1)

    return obj1


def read_json_zstd(obj):
    """
    Deserializer from a compressed zstandard json object to a dictionary.

    Parameters
    ----------
    obj : bytes
        The bytes object.

    Returns
    -------
    Dict
    """
    if isinstance(obj, str):
        with open(obj, 'rb') as p:
            dctx = zstd.ZstdDecompressor()
            with dctx.stream_reader(p) as reader:
                obj1 = reader.read()
    elif isinstance(obj, bytes):
        dctx = zstd.ZstdDecompressor()
        obj1 = dctx.decompress(obj)
    else:
        raise TypeError('obj must either be a str path or a bytes object')

    dict1 = orjson.loads(obj1)

    return dict1


# def s3_client(connection_config: dict, max_pool_connections: int = 30):
#     """
#     Function to establish a client connection with an S3 account. This can use the legacy connect (signature_version s3) and the curent version.

#     Parameters
#     ----------
#     connection_config : dict
#         A dictionary of the connection info necessary to establish an S3 connection. It should contain service_name, endpoint_url, aws_access_key_id, and aws_secret_access_key. connection_config can also be a URL to a public S3 bucket.
#     max_pool_connections : int
#         The number of simultaneous connections for the S3 connection.

#     Returns
#     -------
#     S3 client object
#     """
#     ## Validate config
#     _ = tdm.base.ConnectionConfig(**connection_config)

#     s3_config = copy.deepcopy(connection_config)

#     if 'config' in s3_config:
#         config0 = s3_config.pop('config')
#         config0.update({'max_pool_connections': max_pool_connections})
#         config1 = boto3.session.Config(**config0)

#         s3_config1 = s3_config.copy()
#         s3_config1.update({'config': config1})

#         s3 = boto3.client(**s3_config1)
#     else:
#         s3_config.update({'config': botocore.config.Config(max_pool_connections=max_pool_connections)})
#         s3 = boto3.client(**s3_config)

#     return s3


# def get_object_s3(obj_key: str, bucket: str, s3: botocore.client.BaseClient = None, connection_config: dict = None, public_url: HttpUrl=None, version_id=None, range_start: int=None, range_end: int=None, chunk_size=524288, counter=5):
#     """
#     General function to get an object from an S3 bucket. One of s3, connection_config, or public_url must be used.

#     Parameters
#     ----------
#     obj_key : str
#         The object key in the S3 bucket.
#     s3 : botocore.client.BaseClient
#         An S3 client object created via the s3_client function.
#     connection_config : dict
#         A dictionary of the connection info necessary to establish an S3 connection. It should contain service_name, s3, endpoint_url, aws_access_key_id, and aws_secret_access_key.
#     public_url : str
#         A URL to a public S3 bucket. This is generally only used for Backblaze object storage.
#     bucket : str
#         The bucket name.
#     counter : int
#         Number of times to retry to get the object.

#     Returns
#     -------
#     bytes
#         bytes object of the S3 object.
#     """
#     counter1 = counter

#     transport_params = {'buffer_size': chunk_size}

#     if isinstance(version_id, str):
#         transport_params['version_id'] = version_id

#     ## Headers
#     headers = {}
#     # Range
#     range_dict = {}

#     if range_start is not None:
#         range_dict['start'] = str(range_start)
#     else:
#         range_dict['start'] = ''

#     if range_end is not None:
#         range_dict['end'] = str(range_end)
#     else:
#         range_dict['end'] = ''

#     ## Get the object
#     while True:
#         try:
#             if isinstance(public_url, str) and (version_id is None):
#                 url = create_public_s3_url(public_url, bucket, obj_key)

#                 if range_dict:
#                     range1 = 'bytes={start}-{end}'.format(**range_dict)
#                     headers['Range'] = range1

#                 file_obj = smart_open.open(url, 'rb', headers=headers, transport_params=transport_params)

#             elif isinstance(s3, botocore.client.BaseClient) or isinstance(connection_config, dict):
#                 if range_dict:
#                     range1 = 'bytes={start}-{end}'.format(**range_dict)
#                     transport_params.update({'client_kwargs': {'S3.Client.get_object': {'Range': range1}}})

#                 if s3 is None:
#                     _ = tdm.base.ConnectionConfig(**connection_config)

#                     s3 = s3_client(connection_config)

#                 s3_url = s3_url_base.format(bucket=bucket, key=obj_key)
#                 transport_params['client'] = s3

#                 file_obj = smart_open.open(s3_url, 'rb', transport_params=transport_params)

#             else:
#                 raise TypeError('One of s3, connection_config, or public_url needs to be correctly defined.')

#             break
#         except:
#             # print(traceback.format_exc())
#             if counter1 == 0:
#                 # raise ValueError('Could not properly download the object after several tries')
#                 print('Object could not be downloaded.')
#                 return None
#             else:
#                 # print('Could not properly extract the object; trying again in 5 seconds')
#                 counter1 = counter1 - 1
#                 sleep(3)

#     return file_obj


def chunk_filters(results_chunks, stn_ids, time_interval=None, from_date=None, to_date=None, heights=None, bands=None, from_mod_date=None, to_mod_date=None):
    """

    """
    ## Stations filter
    rc2 = copy.deepcopy([rc for rc in results_chunks if rc['station_id'] in stn_ids])
    first_one = rc2[0]

    ## Temporal filters
    if isinstance(from_date, (str, pd.Timestamp, datetime)) and ('chunk_day' in first_one):
        from_date1 = int(pd.Timestamp(from_date).timestamp()/60/60/24)
        rc2 = [rc for rc in rc2 if (rc['chunk_day'] + time_interval) >= from_date1]

    if len(rc2) == 0:
        return rc2

    if isinstance(to_date, (str, pd.Timestamp, datetime)) and ('chunk_day' in first_one):
        to_date1 = int(pd.Timestamp(to_date).timestamp()/60/60/24)
        rc2 = [rc for rc in rc2 if rc['chunk_day'] <= to_date1]

    if len(rc2) == 0:
        return rc2

    if isinstance(from_mod_date, (str, pd.Timestamp, datetime)) and ('modified_date' in first_one):
        from_mod_date1 = pd.Timestamp(from_mod_date)
        rc2 = [rc for rc in rc2 if pd.Timestamp(rc['modified_date']) >= from_mod_date1]

    if len(rc2) == 0:
        return rc2

    if isinstance(to_mod_date, (str, pd.Timestamp, datetime)) and ('modified_date' in first_one):
        to_mod_date1 = pd.Timestamp(to_mod_date)
        rc2 = [rc for rc in rc2 if pd.Timestamp(rc['modified_date']) <= to_mod_date1]

    if len(rc2) == 0:
        return rc2

    ## Heights and bands filter
    if (heights is not None) and ('height' in first_one):
        if isinstance(heights, (int, float)):
            h1 = [int(heights*1000)]
        elif isinstance(heights, list):
            h1 = [int(h*1000) for h in heights]
        else:
            raise TypeError('heights must be an int, float, or list of int/float.')
        rc2 = [rc for rc in rc2 if rc['height'] in h1]

    if len(rc2) == 0:
        return rc2

    if (bands is not None) and ('band' in first_one):
        if isinstance(bands, int):
            b1 = [heights]
        elif isinstance(bands, list):
            b1 = [int(b) for b in bands]
        else:
            raise TypeError('bands must be an int or list of int.')
        rc2 = [rc for rc in rc2 if rc['band'] in b1]

    if len(rc2) == 0:
        return rc2

    ## Sort by mod date
    rc2.sort(key=lambda d: d['modified_date'] if 'modified_date' in d else '1900-01-01')

    return rc2


# def remove_char_dim_names(data):
#     """

#     """
#     for v in data.variables:
#         if 'char_dim_name' in data[v].encoding:
#             _ = data[v].encoding.pop('char_dim_name')

#     return data


# def remove_results_junk(data):
#     """

#     """
#     chunk_vars = [v for v in list(data.variables) if 'chunk' in v]
#     if chunk_vars:
#         data = data.drop_vars(chunk_vars)

#     # crap_vars = [v for v in list(data.variables) if 'string' in v]
#     # if crap_vars:
#     #     data = data.drop_vars(crap_vars)

#     # if 'geometry' in data.dims:
#     #     stn_vars = [v for v in list(data.data_vars) if ('time' not in data[v].dims) and (v not in ['station_id', 'lon', 'lat'])]
#     #     data = data.drop_vars(stn_vars)

#     if 'station_geometry' in data.dims:
#         stn_vars = [d for d in data.variables if 'station_geometry' in data[d].dims]
#         data = data.drop_vars(stn_vars)

#     return data


def result_filters(h5, from_date=None, to_date=None):
    """

    """
    h5 = h5.sel(exclude_coords=['station_geometry', 'chunk_date'])

    ## Time filters
    if isinstance(from_date, (str, pd.Timestamp, datetime)):
        from_date1 = np.datetime64(from_date)
    else:
        from_date1 = None
    if isinstance(to_date, (str, pd.Timestamp, datetime)):
        to_date1 = np.datetime64(to_date)
    else:
        to_date1 = None

    if (to_date1 is not None) or (from_date1 is not None):
        h5 = h5.sel({'time': slice(from_date1, to_date1)})

    return h5


# def process_results_output(ts_xr, output='xarray', squeeze_dims=False):
#     """

#     """
#     ## Return
#     if squeeze_dims:
#         ts_xr = ts_xr.squeeze()

#     if output == 'xarray':
#         return ts_xr

#     elif output == 'dict':
#         data_dict = ts_xr.to_dict()

#         return data_dict

#     elif output == 'json':
#         json1 = orjson.dumps(ts_xr.to_dict(), option=orjson.OPT_OMIT_MICROSECONDS | orjson.OPT_SERIALIZE_NUMPY)

#         return json1
#     else:
#         raise ValueError("output must be one of 'xarray', 'dict', or 'json'")


# def read_in_chunks(file_object, chunk_size=524288):
#     while True:
#         data = file_object.read(chunk_size)
#         if not data:
#             break
#         yield data


# def local_file_byte_iterator(path, chunk_size=DEFAULT_BUFFER_SIZE):
#     """given a path, return an iterator over the file
#     that lazily loads the file.
#     https://stackoverflow.com/a/37222446/6952674
#     """
#     path = pathlib.Path(path)
#     with path.open('rb') as file:
#         reader = partial(file.read1, DEFAULT_BUFFER_SIZE)
#         file_iterator = iter(reader, bytes())
#         for chunk in file_iterator:
#             yield from chunk


# def stream_to_file(file_obj, file_path, chunk_size=524288):
#     """

#     """
#     file_path1 = pathlib.Path(file_path)
#     file_path1.parent.mkdir(parents=True, exist_ok=True)

#     with open(file_path1, 'wb') as f:
#         chunk = file_obj.read(chunk_size)
#         while chunk:
#             f.write(chunk)
#             chunk = file_obj.read(chunk_size)


# def decompress_stream_to_file(file_obj, file_path, chunk_size=524288):
#     """

#     """
#     file_path1 = pathlib.Path(file_path)
#     file_path1.parent.mkdir(parents=True, exist_ok=True)

#     if file_path1.suffix == '.zst':
#         file_path2 = file_path1.stem
#         dctx = zstd.ZstdDecompressor()

#         with open(file_path2, 'wb') as f:
#             dctx.copy_stream(file_obj, f, read_size=chunk_size, write_size=chunk_size)

#     elif file_path1.suffix == '.gz':
#         file_path2 = file_path1.stem

#         with gzip.open(file_obj, 'rb') as s_file, open(file_path2, 'wb') as d_file:
#             shutil.copyfileobj(s_file, d_file, chunk_size)

#     else:
#         file_path2 = file_path1
#         stream_to_file(file_obj, file_path2, chunk_size)

#     return str(file_path2)


# def decompress_stream_to_object(file_obj, compression, chunk_size=524288):
#     """

#     """
#     b1 = BytesIO()

#     if compression == 'zstd':
#         dctx = zstd.ZstdDecompressor()

#         with open(b1, 'wb') as f:
#             dctx.copy_stream(file_obj, f, read_size=chunk_size, write_size=chunk_size)

#     elif compression == '.gz':

#         with gzip.open(file_obj, 'rb') as s_file, open(b1, 'wb') as d_file:
#             shutil.copyfileobj(s_file, d_file, chunk_size)

#     else:
#         with open(b1, 'wb') as f:
#             chunk = file_obj.read(chunk_size)
#             while chunk:
#                 f.write(chunk)
#                 chunk = file_obj.read(chunk_size)

#     return b1


# def url_stream_to_file(url, file_path, decompress=False, chunk_size=524288):
#     """

#     """
#     file_path1 = pathlib.Path(file_path)
#     if file_path1.is_dir():
#         file_name = url.split('/')[-1]
#         file_path2 = str(file_path1.joinpath(file_name))
#     else:
#         file_path2 = file_path

#     base_path = os.path.split(file_path2)[0]
#     os.makedirs(base_path, exist_ok=True)

#     counter = 4
#     while True:
#         try:
#             with requests.get(url, stream=True, timeout=300) as r:
#                 r.raise_for_status()
#                 stream = ResponseStream(r.iter_content(chunk_size))

#                 if decompress:
#                     if str(file_path2).endswith('.zst'):
#                         file_path2 = os.path.splitext(file_path2)[0]
#                         dctx = zstd.ZstdDecompressor()

#                         with open(file_path2, 'wb') as f:
#                             dctx.copy_stream(stream, f, read_size=chunk_size, write_size=chunk_size)

#                     elif str(file_path2).endswith('.gz'):
#                         file_path2 = os.path.splitext(file_path2)[0]

#                         with gzip.open(stream, 'rb') as s_file, open(file_path2, 'wb') as d_file:
#                             shutil.copyfileobj(s_file, d_file, chunk_size)

#                     else:
#                         with open(file_path2, 'wb') as f:
#                             chunk = stream.read(chunk_size)
#                             while chunk:
#                                 f.write(chunk)
#                                 chunk = stream.read(chunk_size)
#                 else:
#                     with open(file_path2, 'wb') as f:
#                         chunk = stream.read(chunk_size)
#                         while chunk:
#                             f.write(chunk)
#                             chunk = stream.read(chunk_size)

#                 break

#         except Exception as err:
#             if counter < 1:
#                 raise err
#             else:
#                 counter = counter - 1
#                 sleep(5)

#     return file_path2


# def process_dataset(data, from_date=None, to_date=None):
#     """
#     Stupid xarray being inefficient at parsing file objects...
#     """
#     ## Remove junk fields
#     h1 = H5(data).sel(exclude_coords=['station_geometry', 'chunk_date'])

#     h2 = result_filters(h1, from_date, to_date)

#     data_obj = io.BytesIO()
#     h2.to_hdf5(data_obj)

#     return data_obj


def download_results(chunk: dict, bucket: str, s3: botocore.client.BaseClient = None, connection_config: dict = None, public_url: HttpUrl = None, cache: Union[pathlib.Path] = None, from_date=None, to_date=None, return_raw=False):
    """

    """
    file_obj = s3tethys.get_object_s3(chunk['key'], bucket, s3, connection_config, public_url)

    if isinstance(cache, pathlib.Path):
        chunk_hash = chunk['chunk_hash']
        version_date = pd.Timestamp(chunk['version_date']).strftime('%Y%m%d%H%M%SZ')
        results_file_name = local_results_name.format(ds_id=chunk['dataset_id'], stn_id=chunk['station_id'], chunk_id=chunk['chunk_id'], version_date=version_date, chunk_hash=chunk_hash)
        chunk_path = cache.joinpath(results_file_name)

        if not chunk_path.exists():
            chunk_path.parent.mkdir(parents=True, exist_ok=True)

            if chunk['key'].endswith('.zst'):
                data = xr.load_dataset(s3tethys.decompress_stream_to_object(io.BytesIO(file_obj.read()), 'zstd'))
                H5(data).sel(exclude_coords=['station_geometry', 'chunk_date']).to_hdf5(chunk_path, compression='zstd')
                data.close()
                del data
            else:
                s3tethys.stream_to_file(file_obj, chunk_path)

        data_obj = chunk_path

    else:
        if return_raw:
            return file_obj

        if chunk['key'].endswith('.zst'):
            file_obj = s3tethys.decompress_stream_to_object(io.BytesIO(file_obj.read()), 'zstd')
            data = xr.load_dataset(file_obj.read(), engine='scipy')
        else:
            data = io.BytesIO(file_obj.read())

        h1 = H5(data)
        data_obj = io.BytesIO()
        h1 = result_filters(h1)
        h1.to_hdf5(data_obj, compression='zstd')

        if isinstance(data, xr.Dataset):
            data.close()
        del data
        del h1

    del file_obj

    return data_obj


def xr_concat(datasets: List[xr.Dataset]):
    """
    A much more efficient concat/combine of xarray datasets. It's also much safer on memory.
    """
    # Get variables for the creation of blank dataset
    coords_list = []
    chunk_dict = {}

    for chunk in datasets:
        coords_list.append(chunk.coords.to_dataset())
        for var in chunk.data_vars:
            if var not in chunk_dict:
                dims = tuple(chunk[var].dims)
                enc = chunk[var].encoding.copy()
                dtype = chunk[var].dtype
                _ = [enc.pop(d) for d in ['original_shape', 'source'] if d in enc]
                var_dict = {'dims': dims, 'enc': enc, 'dtype': dtype, 'attrs': chunk[var].attrs}
                chunk_dict[var] = var_dict

    try:
        xr3 = xr.combine_by_coords(coords_list, compat='override', data_vars='minimal', coords='all', combine_attrs='override')
    except:
        xr3 = xr.merge(coords_list, compat='override', combine_attrs='override')

    # Run checks - requires psutil which I don't want to make it a dep yet...
    # available_memory = getattr(psutil.virtual_memory(), 'available')
    # dims_dict = dict(xr3.coords.dims)
    # size = 0
    # for var, var_dict in chunk_dict.items():
    #     dims = var_dict['dims']
    #     dtype_size = var_dict['dtype'].itemsize
    #     n_dims = np.prod([dims_dict[dim] for dim in dims])
    #     size = size + (n_dims*dtype_size)

    # if size >= available_memory:
    #     raise MemoryError('Trying to create a dataset of size {}MB, while there is only {}MB available.'.format(int(size*10**-6), int(available_memory*10**-6)))

    # Create the blank dataset
    for var, var_dict in chunk_dict.items():
        dims = var_dict['dims']
        shape = tuple(xr3[c].shape[0] for c in dims)
        xr3[var] = (dims, np.full(shape, np.nan, var_dict['dtype']))
        xr3[var].attrs = var_dict['attrs']
        xr3[var].encoding = var_dict['enc']

    # Fill the dataset with data
    for chunk in datasets:
        for var in chunk.data_vars:
            if isinstance(chunk[var].variable._data, np.ndarray):
                xr3[var].loc[chunk[var].transpose(*chunk_dict[var]['dims']).coords.indexes] = chunk[var].transpose(*chunk_dict[var]['dims']).values
            elif isinstance(chunk[var].variable._data, xr.core.indexing.MemoryCachedArray):
                c1 = chunk[var].copy().load().transpose(*chunk_dict[var]['dims'])
                xr3[var].loc[c1.coords.indexes] = c1.values
                c1.close()
                del c1
            else:
                raise TypeError('Dataset data should be either an ndarray or a MemoryCachedArray.')

    return xr3


def filter_mod_dates(results, from_mod_date=None, to_mod_date=None):
    """
    Need to do this because xarray "where" is useless...
    """
    if ((from_mod_date is not None) or (to_mod_date is not None)) and ('modified_date' in results):
        mod_dates = results['modified_date'].copy().load()

        if (from_mod_date is not None) and (to_mod_date is not None):
            mod_bool = (mod_dates >= pd.Timestamp(from_mod_date)) & (mod_dates <= pd.Timestamp(to_mod_date))
        elif (from_mod_date is not None):
            mod_bool = (mod_dates >= pd.Timestamp(from_mod_date))
        elif (to_mod_date is not None):
            mod_bool = (mod_dates <= pd.Timestamp(to_mod_date))

        data_vars1 = [var for var in results.data_vars if 'time' in results[var].dims]

        results[data_vars1] = results[data_vars1].where(mod_bool)

        return results.dropna('time', how='all')
    else:
        return results


def results_concat(results_list, output_path=None, from_date=None, to_date=None, from_mod_date=None, to_mod_date=None, compression='lzf'):
    """

    """
    if output_path is None:
        output_path = io.BytesIO()
        compression = 'zstd'

    h1 = H5(results_list)
    h1 = result_filters(h1, from_date, to_date)
    h1.to_hdf5(output_path, compression=compression)

    xr3 = xr.open_dataset(output_path, engine='h5netcdf', cache=False)

    ## Deal with mod dates filters
    if ((from_mod_date is not None) or (to_mod_date is not None)) and ('modified_date' in xr3):
        xr3 = filter_mod_dates(xr3, from_mod_date, to_mod_date)

    return xr3







































